<?php

namespace Drupal\txnl_bulk_editing\Plugin\AllowedFieldValues;

use Drupal\Core\Entity\EntityInterface;
use Drupal\txnl_bulk_editing\AllowedValue;
use Ramsey\Uuid\Uuid;

/**
 * Default allowed values handler for drop-down list fields with string values.
 *
 * @AllowedFieldValuesHandler(
 *   id="list_string",
 *   label="List (string)",
 *   description="Handles list fields with pre-configured values.",
 *   field_types={
 *     "list_string"
 *   }
 * )
 */
class StringListFieldHandler extends AllowedFieldValuesPluginBase {

  /**
   * Namespace for list fields that don't have a field definition with a UUID.
   *
   * Configurable content fields are backed by a config entity stored in the
   * database, so we use that UUID as their UUID namespace. We need a UUID
   * namespace for list fields so that we can generate a distinct v5 UUID
   * for each allowed value of the field. The UUID is generated by combining the
   * namespace UUID with the machine name of each value.
   */
  const DEFAULT_FIELD_UUID_NS = '247ecd78-a4da-4434-bb22-d37e8be54b18';

  /**
   * {@inheritdoc}
   */
  public function getAllowedValues(int $start_offset = 0,
                                   int $page_size = 50): array {
    $field_definition = $this->getFieldDefinition();

    $field_allowed_values =
      $field_definition->getSetting('allowed_values') ?? [];

    if ($field_definition instanceof EntityInterface) {
      $field_uuid = $field_definition->uuid();
    }
    else {
      $field_uuid = self::DEFAULT_FIELD_UUID_NS;
    }

    if (!$field_definition->isRequired()) {
      // Add an option to clear the field.
      $static_values = ['' => 'N/A'];
    }
    else {
      $static_values = [];
    }

    $all_allowed_values = array_merge($static_values, $field_allowed_values);

    $ranged_allowed_values =
      array_slice($all_allowed_values, $start_offset, $page_size);

    $referenceable_options = [];

    foreach ($ranged_allowed_values as $key => $label) {
      $option_uuid = Uuid::uuid5($field_uuid, $key)->toString();

      $referenceable_options[] =
        new AllowedValue($option_uuid, $key, $label, $field_definition);
    }

    return $referenceable_options;
  }

}
